from typing import Optional

import numpy as np
from toposort import toposort_flatten, toposort

from sampo.scheduler.base import SchedulerType
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator


class TopologicalScheduler(GenericScheduler):
    """
    Scheduler, that represent 'WorkGraph' in topological order.
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.Topological,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=AverageReqResourceOptimizer(),
                         timeline_type=MomentumTimeline,
                         prioritization_f=self._topological_sort,
                         optimize_resources_f=self.get_default_res_opt_function(lambda _: Time(0)),
                         work_estimator=work_estimator)

    # noinspection PyMethodMayBeStatic
    def _topological_sort(self, wg: WorkGraph, work_estimator: WorkTimeEstimator) -> list[GraphNode]:
        """
        Sort 'WorkGraph' in topological order.

        :param wg: WorkGraph
        :param work_estimator: function that calculates execution time of the work
        :return: list of sorted nodes in graph
        """
        node2ind = {node: i for i, node in enumerate(wg.nodes)}

        dependents: dict[int, set[int]] = {
            node2ind[v]: {node2ind[node] for node in v.parents}
            for v in wg.nodes
        }

        tsorted_nodes_indices: list[int] = toposort_flatten(dependents, sort=True)
        tsorted_nodes = [wg.nodes[i] for i in tsorted_nodes_indices]
        return list(reversed(tsorted_nodes))


class RandomizedTopologicalScheduler(TopologicalScheduler):
    """
    Scheduler, that represent 'WorkGraph' in topological order with random.
    """
    def __init__(self,
                 work_estimator: DefaultWorkEstimator = DefaultWorkEstimator(),
                 random_seed: Optional[int] = None):
        super().__init__(work_estimator=work_estimator)
        self._random_state = np.random.RandomState(random_seed)

    def _topological_sort(self, wg: WorkGraph, work_estimator: WorkTimeEstimator) -> list[GraphNode]:
        def shuffle(nodes: set[GraphNode]) -> list[GraphNode]:
            """
            Shuffle nodes that are on the same level.

            :param nodes: list of nodes
            :return: list of shuffled indices
            """
            nds = list(nodes)
            indices = np.arange(len(nds))
            self._random_state.shuffle(indices)
            return [nds[ind] for ind in indices]

        dependents: dict[GraphNode, set[GraphNode]] = {v: v.parents_set for v in wg.nodes}
        tsorted_nodes: list[GraphNode] = [
            node for level in toposort(dependents)
            for node in shuffle(level)
        ]

        # print([node.work_unit.id for node in tsorted_nodes])
        return list(reversed(tsorted_nodes))



from dataclasses import dataclass

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.requirements import WorkerReq, EquipmentReq, MaterialReq, ConstructionObjectReq
from sampo.schemas.resources import Material
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.utilities.serializers import custom_serializer


@dataclass
class WorkUnit(AutoJSONSerializable['WorkUnit'], Identifiable):
    """
    Class that describe vertex in graph (one work/task)
    """
    def __init__(self, id: str, name: str, worker_reqs: list[WorkerReq] = [], equipment_reqs: list[EquipmentReq] = [],
                 material_reqs: list[MaterialReq] = [], object_reqs: list[ConstructionObjectReq] = [],
                 group: str = 'default', is_service_unit=False, volume: float = 0,
                 volume_type: str = "unit", display_name: str = "", workground_size: int = 100):
        """
        :param worker_reqs: list of required professions (i.e. workers)
        :param equipment_reqs: list of required equipment
        :param material_reqs: list of required materials (e.g. logs, stones, gravel etc.)
        :param object_reqs: list of required objects (e.g. electricity, pipelines, roads)
        :param group: union block of works
        :param is_service_unit: service units are additional vertexes
        :param volume: scope of work
        :param volume_type: unit of scope of work
        :param display_name: name of work
        """
        super(WorkUnit, self).__init__(id, name)
        self.worker_reqs = worker_reqs
        self.equipment_reqs = equipment_reqs
        self.object_reqs = object_reqs
        self.material_reqs = material_reqs
        self.group = group
        self.is_service_unit = is_service_unit
        self.volume = volume
        self.volume_type = volume_type
        self.display_name = display_name
        self.workground_size = workground_size

    def need_materials(self) -> list[Material]:
        return [req.material() for req in self.material_reqs]

    @custom_serializer('worker_reqs')
    def worker_reqs_serializer(self, value: list[WorkerReq]):
        """
        Return serialized list of worker requirements

        :param value: list of worker requirements
        :return: list of worker requirements
        """
        return [wr._serialize() for wr in value]

    @classmethod
    @custom_serializer('worker_reqs', deserializer=True)
    def worker_reqs_deserializer(cls, value):
        """
        Get list of worker requirements

        :param value: serialized list of work requirements
        :return: list of worker requirements
        """
        return [WorkerReq._deserialize(wr) for wr in value]

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        return self._serialize()

    def __setstate__(self, state):
        new_work_unit = self._deserialize(state)
        self.worker_reqs = new_work_unit.worker_reqs
        self.equipment_reqs = new_work_unit.equipment_reqs
        self.object_reqs = new_work_unit.object_reqs
        self.material_reqs = new_work_unit.material_reqs
        self.id = new_work_unit.id
        self.name = new_work_unit.name
        self.is_service_unit = new_work_unit.is_service_unit
        self.volume = new_work_unit.volume
        self.volume_type = new_work_unit.volume_type
        self.group = new_work_unit.group
        self.display_name = new_work_unit.display_name
        self.workground_size = new_work_unit.workground_size


# Function is chosen because it has a quadratic decrease in efficiency as the number of commands on the object
# increases, after the maximum number of commands begins to decrease in efficiency, and its growth rate depends on
# the maximum number of commands.
# sum(1 - ((x-1)^2 / max_groups^2), where x from 1 to groups_count



from random import Random

import sampo.generator.config.gen_counts as gen_c
import sampo.generator.config.worker_req as wr
from sampo.generator.config.worker_req import scale_reqs
from sampo.schemas.graph import GraphNode, EdgeType
from sampo.schemas.interval import IntervalUniform
from sampo.schemas.utils import uuid_str
from sampo.schemas.works import WorkUnit


def _add_addition_work(probability: float, rand: Random | None = None) -> bool:
    """Return answer, if addition work will be added"""
    return IntervalUniform(0, 1).rand_float(rand) <= probability


def _get_roads(parents: list[GraphNode], cluster_name: str, dist: float,
               rand: Random | None = None) -> dict[str, GraphNode]:
    road_nodes = dict()
    min_r = WorkUnit(uuid_str(rand), "minimal road",
                     scale_reqs(wr.MIN_ROAD, dist), group=f"{cluster_name}:road", volume=dist, volume_type="km")
    road_nodes['min'] = GraphNode(min_r, parents)
    temp_r = WorkUnit(uuid_str(rand), "temporary road",
                      scale_reqs(wr.TEMP_ROAD, dist), group=f"{cluster_name}:road", volume=dist, volume_type="km")
    road_nodes['temp'] = GraphNode(temp_r, [(road_nodes['min'], wr.ATOMIC_ROAD_LEN, EdgeType.LagFinishStart)])

    final_r = WorkUnit(uuid_str(rand), "final road", scale_reqs(wr.FINAL_ROAD, dist), group=f"{cluster_name}:road",
                       volume=dist, volume_type="km")
    road_nodes['final'] = GraphNode(final_r, [(road_nodes['temp'], wr.ATOMIC_ROAD_LEN, EdgeType.LagFinishStart)])
    return road_nodes


def _get_engineering_preparation(parents: list[GraphNode], cluster_name: str, boreholes_count: int,
                                 rand: Random | None = None) -> GraphNode:
    worker_req = wr.mul_borehole_volume(wr.ENGINEERING_PREPARATION, boreholes_count, wr.ENGINEERING_PREPARATION_BASE)
    work = WorkUnit(uuid_str(rand), "engineering preparation", worker_req,
                    group=f"{cluster_name}:engineering",
                    volume=wr.get_borehole_volume(boreholes_count, wr.ENGINEERING_PREPARATION_BASE))
    node = GraphNode(work, parents)
    return node


def _get_power_lines(parents: list[GraphNode], cluster_name: str, dist_line: float,
                     dist_high_line: float | None = None, rand: Random | None = None) -> list[GraphNode]:
    worker_req = wr.scale_reqs(wr.POWER_LINE, dist_line)
    power_line_1 = WorkUnit(uuid_str(rand), "power line", worker_req,
                            group=f"{cluster_name}:electricity",
                            volume=dist_line, volume_type="km")
    power_line_2 = WorkUnit(uuid_str(rand), "power line", worker_req,
                            group=f"{cluster_name}:electricity",
                            volume=dist_line, volume_type="km")

    power_lines = [
        GraphNode(power_line_1, parents),
        GraphNode(power_line_2, parents),
    ]
    if dist_high_line is not None:
        worker_req_high = wr.scale_reqs(wr.POWER_LINE, dist_high_line)
        high_power_line = WorkUnit(uuid_str(rand), "high power line", worker_req_high,
                                   group=f"{cluster_name}:electricity", volume=dist_high_line, volume_type="km")
        power_lines.append(GraphNode(high_power_line, parents))

    return power_lines


def _get_pipe_lines(parents: list[GraphNode], cluster_name: str, pipe_dists: list[float],
                    rand: Random | None = None) -> list[GraphNode]:
    worker_req_pipe = wr.scale_reqs(wr.PIPE_LINE, pipe_dists[0])
    first_pipe = WorkUnit(uuid_str(rand), "pipe", worker_req_pipe, group=f"{cluster_name}:oil_gas_long_pipes",
                          volume=pipe_dists[0], volume_type="km")

    graph_nodes = [GraphNode(first_pipe, parents)]
    for i in range(1, len(pipe_dists)):
        node_work = WorkUnit(uuid_str(rand), "node", wr.PIPE_NODE,
                             group=f"{cluster_name}:oil_gas_long_pipes")
        graph_nodes.append(GraphNode(node_work, parents))
        worker_req_pipe = wr.scale_reqs(wr.PIPE_LINE, pipe_dists[i])
        pipe_work = WorkUnit(uuid_str(rand), "pipe", worker_req_pipe,
                             group=f"{cluster_name}:oil_gas_long_pipes",
                             volume=pipe_dists[i], volume_type="km")
        graph_nodes.append(GraphNode(pipe_work, parents))

    worker_req_loop = wr.scale_reqs(wr.PIPE_LINE, pipe_dists[0])
    looping = WorkUnit(uuid_str(rand), "looping", worker_req_loop, group=f"{cluster_name}:oil_gas_long_pipes",
                       volume=pipe_dists[0], volume_type="km")
    graph_nodes.append(GraphNode(looping, graph_nodes[0:1]))
    return graph_nodes


def _get_boreholes_equipment_group(parents: list[GraphNode], cluster_name: str, group_ind: int, borehole_count: int,
                                   rand: Random | None = None) -> list[GraphNode]:
    metering_install = WorkUnit(uuid_str(rand), "metering installation",
                                wr.METERING_INSTALL, group=f"{cluster_name}:borehole_env")
    worker_req_ktp_nep = wr.mul_borehole_volume(wr.KTP_NEP, borehole_count, wr.KTP_NEP_BASE)
    ktp_nep = WorkUnit(uuid_str(rand), "KTP and NEP",
                       worker_req_ktp_nep, group=f"{cluster_name}:borehole_env",
                       volume=wr.get_borehole_volume(borehole_count, wr.KTP_NEP_BASE))
    worker_req_tank = wr.mul_borehole_volume(wr.DRAINAGE_TANK, borehole_count, wr.DRAINAGE_TANK_BASE)
    drainage_tank = WorkUnit(uuid_str(rand), "drainage tank",
                             worker_req_tank, group=f"{cluster_name}:borehole_env",
                             volume=wr.get_borehole_volume(borehole_count, wr.DRAINAGE_TANK_BASE))
    nodes = [
        GraphNode(metering_install, parents),
        GraphNode(ktp_nep, parents),
        GraphNode(drainage_tank, parents),
    ]
    return nodes


def _get_boreholes_equipment_shared(parents: list[GraphNode], cluster_name: str,
                                    rand: Random | None = None) -> list[GraphNode]:
    water_block = WorkUnit(uuid_str(rand), "block water distribution", wr.WATER_BLOCK,
                           group=f"{cluster_name}:borehole_env")
    automation_block = WorkUnit(uuid_str(rand), "block local automation", wr.AUTOMATION_BLOCK,
                                group=f"{cluster_name}:borehole_env")
    block_dosage = WorkUnit(uuid_str(rand), "block dosage inhibitor", wr.BLOCK_DOSAGE,
                            group=f"{cluster_name}:borehole_env")
    start_filters = WorkUnit(uuid_str(rand), "start filters system", wr.START_FILTER,
                             group=f"{cluster_name}:borehole_env")
    firewall = WorkUnit(uuid_str(rand), "firewall tank", wr.FIREWALL, group=f"{cluster_name}:borehole_env")
    nodes = [
        GraphNode(water_block, parents),
        GraphNode(automation_block, parents),
        GraphNode(block_dosage, parents),
        GraphNode(start_filters, parents),
        GraphNode(firewall, parents),
    ]
    return nodes


def _get_boreholes(parents: list[GraphNode], cluster_name: str, group_ind: int, borehole_count: int,
                   rand: Random | None = None) -> list[GraphNode]:
    nodes = []
    for i in range(borehole_count):
        borehole_work = WorkUnit(uuid_str(rand), "borehole",
                                 wr.BOREHOLE, group=f"{cluster_name}:borehole_groups")
        nodes.append(GraphNode(borehole_work, parents))
    return nodes


def _get_boreholes_equipment_general(parents: list[GraphNode], cluster_name: str, pipes_count: int, masts_count: int,
                                     rand: Random | None = None) -> list[GraphNode]:
    nodes = []
    dists_sum = 0
    for i in range(pipes_count):
        dist = gen_c.DIST_BETWEEN_BOREHOLES.rand_float(rand)
        dists_sum += dist
        worker_req_pipe = scale_reqs(wr.POWER_NETWORK, dist)
        pipe_net_work = WorkUnit(uuid_str(rand), "elem of pipe_network", worker_req_pipe,
                                 group=f"{cluster_name}:oil_gas_pipe_net", volume=dist, volume_type="km")
        nodes.append(GraphNode(pipe_net_work, parents))

    worker_req_power = scale_reqs(wr.POWER_NETWORK, dists_sum)
    power_net_work = WorkUnit(uuid_str(rand), "power network", worker_req_power, group=f"{cluster_name}:electricity",
                              volume=dists_sum, volume_type="km")
    nodes.append(GraphNode(power_net_work, parents))

    for i in range(masts_count):
        light_mast_work = WorkUnit(uuid_str(rand), "mast", wr.LIGHT_MAST,
                                   group=f"{cluster_name}:light_masts")
        nodes.append(GraphNode(light_mast_work, parents))
    return nodes


def _get_handing_stage(parents: list[GraphNode], cluster_name: str, borehole_count: int,
                       rand: Random | None = None) -> GraphNode:
    worker_req = wr.mul_borehole_volume(wr.HANDING_STAGE, borehole_count, wr.HANDING_STAGE_BASE)
    work = WorkUnit(uuid_str(rand), "cluster handing", worker_req, group=f"{cluster_name}:handing_stage",
                    volume=wr.get_borehole_volume(borehole_count, wr.HANDING_STAGE_BASE))
    node = GraphNode(work, parents)
    return node


def get_cluster_works(root_node: GraphNode, cluster_name: str, pipe_nodes_count: int,
                      pipe_net_count: int, light_masts_count: int, borehole_counts: list[int],
                      roads: dict[str, GraphNode] | None = None,
                      rand: Random | None = None) -> (GraphNode, dict[str, GraphNode]):
    """
    Creates works on the development of the field on one object, i.e. a group of boreholes

    :param root_node: The parent node of the work graph for this object
    :param cluster_name: object name
    :param pipe_nodes_count: Number of pipeline segments from this field to other fields
    :param pipe_net_count: Number of pipes connecting boreholes
    :param light_masts_count: Number of floodlight masts
    :param borehole_counts: Number of boreholes
    :param roads: If the object is not connected to the central node, the road to which it is connected, otherwise None
    :param rand: Number generator with a fixed seed, or None for no fixed seed
    :return:
    """
    is_slave = roads is not None
    dist_to_parent = None
    if not is_slave:
        dist_to_parent = gen_c.DIST_TO_PARENT.rand_float(rand)
        roads = _get_roads([root_node], cluster_name, dist_to_parent, rand)
    engineering_preparation = _get_engineering_preparation([roads['min']], cluster_name, sum(borehole_counts), rand)
    preparation_stage = [roads['temp'], engineering_preparation]

    dist_line = gen_c.ONE_SECTION_PIPE.rand_float(rand)
    power_lines = _get_power_lines(preparation_stage, cluster_name, dist_line, dist_high_line=dist_to_parent, rand=rand)

    pipe_dists = [gen_c.ONE_SECTION_PIPE.rand_float(rand) for _ in range(pipe_nodes_count + 1)]
    pipe_lines_and_nodes = _get_pipe_lines(preparation_stage, cluster_name, pipe_dists, rand)

    boreholes_eq_shared = _get_boreholes_equipment_shared(preparation_stage, cluster_name, rand)
    boreholes_eq_general = _get_boreholes_equipment_general(preparation_stage, cluster_name,
                                                            pipes_count=pipe_net_count, masts_count=light_masts_count,
                                                            rand=rand)
    boreholes_all = []
    for i, borehole_count in enumerate(borehole_counts):
        boreholes_eq = _get_boreholes_equipment_group(preparation_stage, cluster_name,
                                                      group_ind=i, borehole_count=borehole_count, rand=rand)
        boreholes = _get_boreholes(boreholes_eq + boreholes_eq_shared, cluster_name,
                                   group_ind=i, borehole_count=borehole_count, rand=rand)
        boreholes_all += boreholes

    all_stages = [roads['final']] + power_lines + pipe_lines_and_nodes + boreholes_all + boreholes_eq_general
    handing_stage = _get_handing_stage(all_stages, cluster_name, sum(borehole_counts), rand=rand)
    return handing_stage, roads



from copy import deepcopy
from dataclasses import dataclass
from typing import Any

from sampo.schemas.contractor import Contractor
from sampo.schemas.landscape import MaterialDelivery
from sampo.schemas.resources import Equipment, ConstructionObject, Worker
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.works import WorkUnit
from sampo.utilities.serializers import custom_serializer


@dataclass
class ScheduledWork(AutoJSONSerializable['ScheduledWork']):
    """
    Contains all neccessary info to represent WorkUnit in Scheduler:

    * WorkUnit
    * list of workers, that are required to complete task
    * start and end time
    * contractor, that complete task
    * list of equipment, that is needed to complete the task
    * list of materials - set of non-renewable resources
    * object - variable, that is used in landscape
    """

    ignored_fields = ['equipments', 'materials', 'object']

    def __init__(self,
                 work_unit: WorkUnit,
                 start_end_time: tuple[Time, Time],
                 workers: list[Worker],
                 contractor: Contractor | str,
                 equipments: list[Equipment] | None = None,
                 materials: list[MaterialDelivery] | None = None,
                 object: ConstructionObject | None = None):
        self.work_unit = work_unit
        self.start_end_time = start_end_time
        self.workers = workers
        self.equipments = equipments
        self.materials = materials
        self.object = object

        if contractor is not None:
            if isinstance(contractor, str):
                self.contractor = contractor
            else:
                self.contractor = contractor.name if contractor.name else contractor.id
        else:
            self.contractor = ""

        self.cost = 0
        for worker in self.workers:
            self.cost += worker.get_cost() * self.duration.value

    def __str__(self):
        return f'ScheduledWork[work_unit={self.work_unit}, start_end_time={self.start_end_time}, ' \
               f'workers={self.workers}, contractor={self.contractor}]'

    def __repr__(self):
        return self.__str__()

    @custom_serializer('workers')
    @custom_serializer('start_end_time')
    def serialize_serializable_list(self, value):
        return [t._serialize() for t in value]

    @classmethod
    @custom_serializer('start_end_time', deserializer=True)
    def deserialize_time(cls, value):
        return [Time._deserialize(t) for t in value]

    @classmethod
    @custom_serializer('workers', deserializer=True)
    def deserialize_workers(cls, value):
        return [Worker._deserialize(t) for t in value]

    def get_actual_duration(self, work_estimator: WorkTimeEstimator) -> Time:
        return work_estimator.estimate_time(self.work_unit, self.workers)

    @property
    def start_time(self) -> Time:
        return self.start_end_time[0]

    @start_time.setter
    def start_time(self, val: Time):
        self.start_end_time = (val, self.start_end_time[1])

    @property
    def finish_time(self) -> Time:
        return self.start_end_time[1]

    @property
    def min_child_start_time(self) -> Time:
        return self.finish_time if self.work_unit.is_service_unit else self.finish_time + 1

    @finish_time.setter
    def finish_time(self, val: Time):
        self.start_end_time = (self.start_end_time[0], val)

    @staticmethod
    def start_time_getter():
        return lambda x: x.start_end_time[0]

    @staticmethod
    def finish_time_getter():
        return lambda x: x.start_end_time[1]

    @property
    def duration(self) -> Time:
        start, end = self.start_end_time
        return end - start
    
    def is_overlapped(self, time: int) -> bool:
        start, end = self.start_end_time
        return start <= time < end

    def to_dict(self) -> dict[str, Any]:
        return {
            "task_id": self.work_unit.id,
            "task_name": self.work_unit.name,
            "start": self.start_time.value,
            "finish": self.finish_time.value,
            "contractor_id": self.contractor,
            "workers": {worker.name: worker.count for worker in self.workers},
        }

    def __deepcopy__(self, memodict={}):
        return ScheduledWork(deepcopy(self.work_unit, memodict),
                             deepcopy(self.start_end_time, memodict),
                             deepcopy(self.workers, memodict),
                             self.contractor)



from sampo.schemas.schedule import Schedule


def schedule_cost(schedule: Schedule) -> float:
    """
    Count the summary cost of received schedule

    :param schedule:
    :return:
    """
    cost: float = 0

    for work in schedule.works:
        cost += work.cost
    return cost



import hashlib
import queue
from collections import defaultdict
from typing import Optional, Callable

import pandas as pd
from matplotlib import pyplot as plt, patches, axes
from matplotlib.figure import Figure
from matplotlib.patches import Patch

from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.works import WorkUnit

X_STEP = 0.8
X_DELTA = 0.2
X_PERIOD = X_STEP + X_DELTA

TEXT_X_DELTA = 0.01
TEXT_Y_DELTA = 0.05

Y_STEP = 0.7
Y_DELTA = 0.3
Y_PERIOD = Y_STEP + Y_DELTA

MIN_LEN = 0.2
ARROW_HEAD_WIDTH = Y_STEP / 2
ARROW_HEAD_LENGTH = X_DELTA / 10
BORDER_LINE_WIDTH = 0.3
LINE_WIDTH = 0.6

INF_INT = int(1e9)

SIZE_LIMIT = int(1e3)
DEFAULT_DPI = 50
DPI_LIMIT = 3e5 / SIZE_LIMIT


def work_graph_fig(graph: WorkGraph or GraphNode, fig_size: tuple[int, int],
                   fig_dpi: Optional[int] = 300,
                   max_deep: Optional[int] = None,
                   show_names: Optional[bool] = False, show_arrows: Optional[bool] = True,
                   hide_node_ids: Optional[list[str]] = None,
                   legend_shift: Optional[int] = 0,
                   show_only_not_dotted: Optional[bool] = False,
                   dotted_edges: set[tuple[str, str]] = None,
                   black_list_edges: set[tuple[str, str]] = None,
                   jobs2text_function: Optional[Callable[[pd.Series], str]] = None,
                   text_size: Optional[int] = 1) -> Figure:
    start = graph.start if isinstance(graph, WorkGraph) else graph
    jobs, id_to_job, _ = collect_jobs(start, max_deep)
    jobs = setup_jobs(start, jobs, id_to_job)
    dotted_edges = dotted_edges or set()
    black_list_edges = black_list_edges or set()
    hide_node_ids = set(hide_node_ids or [])

    df = pd.DataFrame(jobs)
    df = df.sort_values(by=["cluster", "group"])

    fig, ax = plt.subplots(1, figsize=fig_size, dpi=fig_dpi)
    ax_add_works(ax, df, show_names, hide_node_ids, jobs2text_function, text_size)
    if show_arrows:
        ax_add_dependencies(ax, df, id_to_job, dotted_edges, hide_node_ids, show_only_not_dotted, black_list_edges)

    ax.set_xlim(df.start.min() - X_DELTA, df.finish.max() + X_DELTA + legend_shift)
    ax.set_ylim(df.y_position.min() - Y_PERIOD, df.y_position.max() + 2 * Y_PERIOD)

    groups = df.group.unique()
    legend_elements = [Patch(facecolor=color_from_str(g_name), label=g_name) for g_name in groups]
    plt.legend(handles=legend_elements, loc='upper right')
    return fig


def extract_cluster_name(work_name: str) -> str:
    cluster_name = work_name.split("`")
    cluster_name = "" if len(cluster_name) < 2 else cluster_name[1]
    return cluster_name


def calculate_work_volume(work_unit: WorkUnit) -> float:
    return sum((req.volume for req in work_unit.worker_reqs))


def collect_jobs(start: GraphNode, max_deep: Optional[int] = None) -> (list[dict], dict[str, int], dict[str, str]):
    max_deep = max_deep or INF_INT
    q = queue.Queue()
    q.put((0, start))
    id_to_job: dict[str, int] = dict()
    used: set[str] = {start.id}
    jobs: list[dict] = []
    colors: dict[str, str] = dict()
    max_volume: float = 0

    while not q.empty():
        deep, node = q.get()
        if deep >= max_deep:
            continue
        unit = node.work_unit
        volume = calculate_work_volume(unit)
        max_volume = max(max_volume, volume)
        id_to_job[unit.id] = len(jobs)
        jobs.append(
            dict(job_id=len(jobs), work_id=str(unit.id), task=unit.name, start=X_PERIOD * deep,
                 children=node.children, parents=node.parents,
                 group=unit.group, color=color_from_str(unit.group), volume=volume,
                 cluster=extract_cluster_name(unit.name)))
        colors[unit.group] = color_from_str(unit.group)
        for child in node.children:
            work_id = child.id
            if work_id in used:
                continue
            q.put((deep + 1, child))
            used.add(work_id)
    return jobs, id_to_job, colors


def setup_jobs(start: GraphNode, jobs: list[dict], id_to_job: dict[str, int]) -> list[dict]:
    cluster_deep_counts: dict[str, dict[int, int]] = defaultdict(lambda: defaultdict(int))

    not_used_parents = {job["job_id"]: len(job["parents"]) for job in jobs}

    q = queue.Queue()
    start_ind = id_to_job[start.id]
    q.put((0, start_ind))

    while not q.empty():
        deep, ind = q.get()
        jobs[ind]["start"] = deep * X_PERIOD
        jobs[ind]["finish"] = jobs[ind]["start"] + X_STEP
        cluster_deep_counts[jobs[ind]["cluster"]][jobs[ind]["start"]] += 1

        for node in jobs[ind]["children"]:
            if node.id not in id_to_job:
                continue
            child_ind = id_to_job[node.id]
            not_used_parents[child_ind] -= 1
            if not_used_parents[child_ind] <= 0:
                q.put((deep + 1, child_ind))

    max_y_pos: dict[int, int] = dict()

    def dfs(v_ind: int, y_pos: int) -> int:
        jobs[v_ind]["y_position"] = y_pos

        used = [id_to_job[child.id] for child in jobs[v_ind]['children'] if id_to_job[child.id] in max_y_pos]
        y_pos = max([y_pos] + [max_y_pos[c_ind] for c_ind in used] or [y_pos])

        for child in jobs[v_ind]['children']:
            c_ind = id_to_job[child.id]
            if c_ind not in max_y_pos:
                y_pos = dfs(c_ind, y_pos) + 1

        max_y_pos[v_ind] = max(jobs[v_ind]["y_position"], y_pos - 1)
        return max_y_pos[v_ind]

    _ = dfs(start_ind, 0)
    return jobs


def ax_add_works(ax: axes.Axes, df: pd.DataFrame, show_names: bool, hide_nodes_id: set[str],
                 jobs2text_function: Optional[Callable[[pd.Series], str]] = None, text_size: Optional[int] = 1):
    if jobs2text_function is None:
        jobs2text_function = default_job2text
    if not show_names:
        jobs2text_function = empty_job2text

    for ind, row in df.iterrows():
        if row.work_id in hide_nodes_id:
            continue
        x: float = float(row.start)
        y: float = float(row.y_position)
        length: float = row.finish - row.start
        rect = patches.Rectangle((x, y), length, Y_STEP, linewidth=BORDER_LINE_WIDTH, edgecolor="k",
                                 facecolor=row.color)
        ax.add_patch(rect)
        ax.text(x + TEXT_X_DELTA, y + TEXT_Y_DELTA, jobs2text_function(row), fontsize=text_size)


def ax_add_dependencies(ax: axes.Axes, df: pd.DataFrame, id_to_job: dict[str, int],
                        dotted_edges: set[tuple[str, str]], hide_nodes_id: set[str], show_only_not_dotted: bool,
                        black_list_edges: set[tuple[str, str]]):
    for ind, job in df.iterrows():
        if job.work_id in hide_nodes_id:
            continue
        for son in job.children:
            if son.id in hide_nodes_id or son.id not in id_to_job:
                continue
            linestyle = 'dotted' if ((str(job.work_id), str(son.id)) in dotted_edges) else None
            if linestyle == 'dotted' and show_only_not_dotted or (str(job.work_id), str(son.id)) in black_list_edges:
                continue
            draw_arrow_between_jobs(ax, job, df[df.job_id == id_to_job[son.id]].iloc[0], linestyle=linestyle)


def draw_arrow_between_jobs(ax, first_job_dict, second_job_dict, linestyle: Optional[str] = None):
    x_fj = first_job_dict.finish
    y_fj = first_job_dict.y_position + Y_STEP / 2
    x_mid = second_job_dict.start - X_DELTA / 2

    x_sj = second_job_dict.start
    y_sj = second_job_dict.y_position + Y_STEP / 2
    line_color = middle_color(first_job_dict["color"], second_job_dict['color'])

    ax.plot([x_fj, x_mid, x_mid], [y_fj, y_fj, y_sj], color=line_color, linewidth=LINE_WIDTH, linestyle=linestyle)
    ax.arrow(x_mid, y_sj, x_sj - x_mid - ARROW_HEAD_LENGTH, 0, color=line_color, linewidth=LINE_WIDTH, shape="full",
             head_width=ARROW_HEAD_WIDTH,
             head_length=ARROW_HEAD_LENGTH, linestyle=linestyle)
    return


def color_from_str(name: str) -> str:
    name = name or ''
    hashed_string = hashlib.sha256(name.encode())
    rgb_code = hashed_string.hexdigest()[len(hashed_string.hexdigest()) - 6:]
    return f"#{rgb_code}"


def middle_color(color_a: str, color_b: str) -> str:
    mid = "#"
    for i in range(1, 7, 2):
        a = int(color_a[i:i + 2], 16)
        b = int(color_b[i:i + 2], 16)
        c = int((a + b) / 2 * 0.7)
        mid += "%0.2X" % c
    return mid


def default_job2text(row: pd.Series) -> str:
    return row.work_id + " " + row.task


def empty_job2text(row: pd.Series) -> str:
    return ''



from datetime import timedelta
from typing import Optional

import pandas as pd
import plotly.express as px

from sampo.utilities.visualization.base import VisualizationMode, visualize


def schedule_gant_chart_fig(schedule_dataframe: pd.DataFrame,
                            visualization: VisualizationMode,
                            remove_service_tasks: bool = False,
                            fig_file_name: Optional[str] = None):
    """
    Creates and saves a gant chart of the scheduled tasks to the specified path.

    :param fig_file_name:
    :param visualization:
    :param remove_service_tasks:
    :param schedule_dataframe: Pandas DataFrame with the information about schedule
    """
    if remove_service_tasks:
        schedule_dataframe = schedule_dataframe.loc[~schedule_dataframe.loc[:, 'task_name'].str.contains('марке')]
        schedule_dataframe = schedule_dataframe.loc[schedule_dataframe.loc[:, 'volume'] >= 0.1]

    schedule_dataframe = schedule_dataframe.rename({'workers': 'workers_dict'}, axis=1)
    schedule_dataframe.loc[:, 'workers'] = schedule_dataframe.loc[:, 'workers_dict']\
        .apply(lambda x: x.replace(", '", ", <br>'"))

    schedule_start = schedule_dataframe.loc[:, 'start'].min()
    schedule_finish = schedule_dataframe.loc[:, 'finish'].max()
    visualization_start_delta = timedelta(days=2)
    visualization_finish_delta = timedelta(days=(schedule_finish - schedule_start).days // 3)

    fig = px.timeline(schedule_dataframe, x_start='start', x_end='finish', y='idx', hover_name='task_name',
                      color=schedule_dataframe.loc[:, 'contractor'],
                      hover_data=['task_name_mapped', 'cost', 'volume', 'measurement', 'workers'],
                      title=f"{'Project tasks - Gant chart'}",
                      category_orders={'idx': list(schedule_dataframe.idx)},
                      text='task_name')

    fig.update_traces(textposition='outside')

    fig.update_yaxes(showticklabels=False, title_text='Project tasks',
                     type="category")

    fig.update_xaxes(type="date",
                     range=[schedule_start - visualization_start_delta,
                            schedule_finish + visualization_finish_delta],
                     title_text='Date')

    fig.update_layout(autosize=True, font_size=12)

    return visualize(fig, mode=visualization, file_name=fig_file_name)



import math
from operator import itemgetter
from typing import Callable

from sampo.schemas.exceptions import NotEnoughMaterialsInDepots, NoAvailableResources
from sampo.schemas.landscape import LandscapeConfiguration, MaterialDelivery
from sampo.schemas.resources import Material
from sampo.schemas.sorted_list import ExtendedSortedList
from sampo.schemas.time import Time


class SupplyTimeline:
    def __init__(self, landscape_config: LandscapeConfiguration):
        self._timeline = {}
        self._capacity = {}
        # material -> list of depots, that can supply this type of resource
        self._resource_sources: dict[str, dict[str, int]] = {}
        for landscape in landscape_config.get_all_resources():
            self._timeline[landscape.id] = ExtendedSortedList([(Time(0), landscape.count), (Time.inf(), 0)],
                                                              itemgetter(0))
            self._capacity[landscape.id] = landscape.count
            for count, res in landscape.get_available_resources():
                res_source = self._resource_sources.get(res, None)
                if res_source is None:
                    res_source = {}
                    self._resource_sources[res] = res_source
                res_source[landscape.id] = count

    def find_min_material_time(self, id: str, start_time: Time, materials: list[Material], batch_size: int) -> Time:
        sum_materials = sum([material.count for material in materials])
        ratio = sum_materials / batch_size
        batches = math.ceil(ratio)

        first_batch = [material.copy().with_count(material.count // batches) for material in materials]
        return self.supply_resources(id, start_time, first_batch, True)[1]

    def deliver_materials(self, id: str, start_time: Time, finish_time: Time,
                          materials: list[Material], batch_size: int) -> tuple[list[MaterialDelivery], Time, Time]:
        """
        Models material delivery.

        Delivery performed in batches sized by batch_size.

        :return: pair of material-driven minimum start and finish times
        """
        sum_materials = sum([material.count for material in materials])
        ratio = sum_materials / batch_size
        batches = math.ceil(ratio)

        first_batch = [material.copy().with_count(material.count // batches) for material in materials]
        other_batches = [first_batch for _ in range(batches - 1)]
        other_batches.append([material.copy().with_count(material.count * (ratio - batches)) for material in materials])

        deliveries = []
        d, start_time = self.supply_resources(id, start_time, first_batch, False)
        deliveries.append(d)
        batch_processing = [self.supply_resources(id, finish_time, batch, False) for batch in other_batches]
        finish_time = max([b[1] for b in batch_processing])
        deliveries.extend([b[0] for b in batch_processing])

        return deliveries, start_time, finish_time

    def _find_best_supply(self, material: str, count: int) -> str:
        # TODO Make better algorithm
        # Return the first depot that can supply given materials
        if len(self._resource_sources) == 0:
            raise NoAvailableResources(f"Schedule can not be built. No available resource sources with material {material}")
        depots = [depot_id for depot_id, depot_count in self._resource_sources[material].items()
                  if depot_count >= count]
        if len(depots) == 0:
            raise NotEnoughMaterialsInDepots(f"Schedule can not be built. No one supplier has enough '{material}' material")
        return depots[0]

    @staticmethod
    def _grab_from_current_area(material_timeline: ExtendedSortedList,
                                cur_time: Time, idx_start: int, need_count: int,
                                capacity: int, going_right: bool, simulate: bool,
                                delivery_writer: Callable[[Time, int], None]) -> Time:
        """
        Processes the whole area starts with `idx_start` from `cur_time` moment

        :param material_timeline:
        :param cur_time:
        :param idx_start:
        :param need_count:
        :param capacity:
        :param going_right:
        :param simulate:
        :return: pair of finish time and grabbed amount
        """
        time_start = material_timeline[idx_start][0]
        time_end = material_timeline[idx_start + 1][0]

        def process_start_milestone():
            nonlocal cur_time, need_count, going_right
            if cur_time == time_start:  # grab from start milestone
                start_count = material_timeline[idx_start][1]
                if need_count > start_count:
                    need_count -= start_count
                    if not simulate:  # drop start milestone
                        material_timeline[idx_start] = (time_start, 0)
                    delivery_writer(cur_time, start_count)  # write to the result

                    if going_right:
                        cur_time += 1
                    else:
                        cur_time -= 1
                else:
                    if not simulate:  # subtract from start milestone
                        material_timeline[idx_start] = (time_start, start_count - need_count)
                    delivery_writer(cur_time, need_count)  # write to the result
                    need_count = 0

        process_start_milestone()

        # TODO Check that we can remove this
        # if not going_right:
        #     return cur_time

        # grab from area
        if going_right:
            while need_count > 0 and cur_time < time_end:  # inside area
                delivery_writer(cur_time, capacity)  # write to the result
                need_count -= capacity
                cur_time += 1

            if cur_time >= time_end:
                # we grabbed all the area, so we don't need to insert any milestone
                # our start milestone is already processed upper
                return cur_time

            if not simulate:
                # if we are here, need_count < 0
                # we grabbed not all the area, so insert milestone to cur_time - 1 moment.
                # 'cur_time - 1' because cur_time is the moment after the last 'need_count' addition performed
                # '-need_count' is resources count that are left at the last seen time moment
                material_timeline.add((cur_time - 1, -need_count))
        else:
            while need_count > 0 and time_start < cur_time:  # inside area
                delivery_writer(cur_time, capacity)  # write to the result
                need_count -= capacity
                cur_time -= 1

            if cur_time == time_start:
                # we grabbed all the area, so now we are allowed to grab the start milestone
                process_start_milestone()
                return cur_time

            if not simulate:
                # if we are here, need_count < 0
                # we grabbed not all the area, so insert milestone to cur_time + 1 moment.
                # 'cur_time + 1' because cur_time is the moment after the last 'need_count' subtraction performed
                # '-need_count' is resources count that are left at the last seen time moment
                material_timeline.add((cur_time + 1, -need_count))

        return cur_time

    def supply_resources(self, id: str, deadline: Time, materials: list[Material], simulate: bool) \
            -> tuple[MaterialDelivery, Time]:
        """
        Finds minimal time that given materials can be supplied, greater that given start time

        :param id: work id
        :param deadline: the time work starts
        :param materials: material resources that are required to start
        :param simulate: should timeline only find minimum supply time and not change timeline
        # :param workground_size: material capacity of the workground
        :return: the time when resources are ready
        """
        min_start_time = deadline
        delivery = MaterialDelivery(id)

        grabbed = 0

        cur_start_time = deadline
        going_right = False

        for material in materials:
            depot = self._find_best_supply(material.name, material.count)
            material_timeline = self._timeline[depot]
            capacity = self._capacity[depot]

            def record_delivery(time: Time, count: int):
                nonlocal count_left
                count_left -= count

                if not simulate:
                    # update depot state
                    self._resource_sources[material.name][depot] -= count
                    # add to the result
                    delivery.add_delivery(material.name, time, count)

            count_left = material.count
            cur_start_time = deadline

            going_right = False

            i = 0
            while count_left > 0:
                if i > 0 and i % 50 == 0:
                    print(f'Probably cycle: cur_start_time={cur_start_time}, idx_left={idx_left}', flush=True)

                # find current area
                idx_left = material_timeline.bisect_key_right(cur_start_time) - 1
                time_left = material_timeline[idx_left][0]
                time_right = material_timeline[idx_left + 1][0]

                cur_start_time = self._grab_from_current_area(material_timeline, cur_start_time, idx_left,
                                                              count_left, capacity, going_right, simulate,
                                                              record_delivery)

                # record_delivery(material.name, depot, cur_start_time, material.count)

                if cur_start_time < time_left:
                    # over left of current area, step left
                    idx_left -= 1
                    if idx_left < 0:
                        cur_start_time = deadline
                        going_right = True
                elif cur_start_time > time_right:
                    # over right of current area, step right
                    idx_left += 1

                i += 1

            min_start_time = min(min_start_time, cur_start_time)

        return delivery, min_start_time

    @property
    def resource_sources(self):
        return self._resource_sources



from sampo.structurator.base import graph_restructuring, STAGE_SEP
from sampo.structurator.graph_insertion import graph_in_graph_insertion
from sampo.structurator.light_modifications import work_graph_ids_simplification



import math
import random
import time
from typing import Callable

import numpy as np
import seaborn as sns
from deap import tools
from deap.tools import initRepeat
from matplotlib import pyplot as plt
from pandas import DataFrame

from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome
from sampo.scheduler.genetic.operators import init_toolbox, ChromosomeType, Individual, copy_chromosome, \
    FitnessFunction, TimeFitness, is_chromosome_correct
from sampo.scheduler.native_wrapper import NativeWrapper
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.exceptions import NoSufficientContractorError
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.schedule import ScheduleWorkDict, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.collections_util import reverse_dictionary


def build_schedule(wg: WorkGraph,
                   contractors: list[Contractor],
                   worker_pool: WorkerContractorPool,
                   population_size: int,
                   generation_number: int,
                   selection_size: int,
                   mutate_order: float,
                   mutate_resources: float,
                   init_schedules: dict[str, tuple[Schedule, list[GraphNode] | None]],
                   rand: random.Random,
                   spec: ScheduleSpec,
                   landscape: LandscapeConfiguration = LandscapeConfiguration(),
                   fitness_constructor: Callable[[Callable[[list[ChromosomeType]], list[int]]],
                                                 FitnessFunction] = TimeFitness,
                   work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                   show_fitness_graph: bool = False,
                   n_cpu: int = 1,
                   assigned_parent_time: Time = Time(0),
                   timeline: Timeline | None = None,
                   time_border: int = None) \
        -> tuple[ScheduleWorkDict, Time, Timeline, list[GraphNode]]:
    """
    Genetic algorithm.
    Structure of chromosome:
    [[order of job], [numbers of workers types 1 for each job], [numbers of workers types 2], ... ]

    Different mate and mutation for order and for workers.
    Generate order of job by prioritization from HEFTs and from Topological.
    Generate resources from min to max.
    Overall initial population is valid.

    :param spec: spec for current scheduling
    :param n_cpu: number or parallel workers to use in computational process
    :param assigned_parent_time: start time of the whole schedule(time shift)
    :return: scheduler
    """

    if show_fitness_graph:
        fitness_history = []

    global_start = time.time()

    start = time.time()
    # preparing access-optimized data structures
    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    index2node: dict[int, GraphNode] = dict(enumerate(nodes))
    work_id2index: dict[str, int] = {node.id: index for index, node in index2node.items()}
    worker_name2index = {worker_name: index for index, worker_name in enumerate(worker_pool)}
    index2contractor = {ind: contractor.id for ind, contractor in enumerate(contractors)}
    index2contractor_obj = dict(enumerate(contractors))
    contractor2index = reverse_dictionary(index2contractor)
    worker_pool_indices = {worker_name2index[worker_name]: {
        contractor2index[contractor_id]: worker for contractor_id, worker in workers_of_type.items()
    } for worker_name, workers_of_type in worker_pool.items()}
    node_indices = list(range(len(nodes)))

    contractors_capacity = np.zeros((len(contractors), len(worker_pool)))
    for w_ind, cont2worker in worker_pool_indices.items():
        for c_ind, worker in cont2worker.items():
            contractors_capacity[c_ind][w_ind] = worker.count

    resources_border = np.zeros((2, len(worker_pool), len(index2node)))
    resources_min_border = np.zeros((len(worker_pool)))
    for work_index, node in index2node.items():
        for req in node.work_unit.worker_reqs:
            worker_index = worker_name2index[req.kind]
            resources_border[0, worker_index, work_index] = req.min_count
            resources_border[1, worker_index, work_index] = req.max_count
            resources_min_border[worker_index] = max(resources_min_border[worker_index], req.min_count)

    contractor_borders = np.zeros((len(contractor2index), len(worker_name2index)), dtype=int)
    for ind, contractor in enumerate(contractors):
        for ind_worker, worker in enumerate(contractor.workers.values()):
            contractor_borders[ind, ind_worker] = worker.count

    # construct inseparable_child -> inseparable_parent mapping
    inseparable_parents = {}
    for node in nodes:
        for child in node.get_inseparable_chain_with_self():
            inseparable_parents[child] = node

    # here we aggregate information about relationships from the whole inseparable chain
    children = {work_id2index[node.id]: list({work_id2index[inseparable_parents[child].id]
                                                  for inseparable in node.get_inseparable_chain_with_self()
                                                  for child in inseparable.children})
                for node in nodes}

    parents = {work_id2index[node.id]: [] for node in nodes}
    for node, node_children in children.items():
        for child in node_children:
            parents[child].append(node)

    print(f'Genetic optimizing took {(time.time() - start) * 1000} ms')

    start = time.time()

    # initial chromosomes construction
    init_chromosomes: dict[str, ChromosomeType] = \
        {name: convert_schedule_to_chromosome(wg, work_id2index, worker_name2index,
                                              contractor2index, contractor_borders, schedule, order)
            if schedule is not None else None
         for name, (schedule, order) in init_schedules.items()}

    toolbox = init_toolbox(wg, contractors, worker_pool, landscape, index2node,
                           work_id2index, worker_name2index, index2contractor,
                           index2contractor_obj, init_chromosomes, mutate_order,
                           mutate_resources, selection_size, rand, spec, worker_pool_indices,
                           contractor2index, contractor_borders, node_indices, parents,
                           assigned_parent_time, work_estimator)

    for name, chromosome in init_chromosomes.items():
        if chromosome is not None:
            if not is_chromosome_correct(chromosome, node_indices, parents):
                raise NoSufficientContractorError('HEFTs are deploying wrong chromosomes')

    native = NativeWrapper(toolbox, wg, contractors, worker_name2index, worker_pool_indices,
                           parents, work_estimator)
    # create population of a given size
    pop = toolbox.population(n=population_size)

    print(f'Toolbox initialization & first population took {(time.time() - start) * 1000} ms')

    if not native.native:
        # save best individuals
        hof = tools.HallOfFame(1, similar=compare_individuals)

        # probability to participate in mutation and crossover for each individual
        cxpb, mutpb = mutate_order, mutate_order
        mutpb_res, cxpb_res = mutate_resources, mutate_resources

        # def evaluate_chromosomes(chromosomes: list[ChromosomeType]):
        #     return native.evaluate([chromo for chromo in chromosomes if toolbox.validate(chromo)])

        fitness_f = fitness_constructor(native.evaluate)

        start = time.time()

        # map to each individual fitness function
        pop = [ind for ind in pop if toolbox.validate(ind[0])]
        fitness = fitness_f.evaluate([ind[0] for ind in pop])

        evaluation_time = time.time() - start

        for ind, fit in zip(pop, fitness):
            ind.fitness.values = [fit]

        hof.update(pop)
        best_fitness = hof[0].fitness.values[0]

        if show_fitness_graph:
            fitness_history.append(sum(fitness) / len(fitness))

        generation = 0
        # the best fitness, track to increase performance by stopping evaluation when not decreasing
        prev_best_fitness = Time.inf()

        print(f'First population evaluation took {(time.time() - start) * 1000} ms')
        start = time.time()

        plateau_steps = 0
        max_plateau_steps = generation_number  # 8

        while generation < generation_number and plateau_steps < max_plateau_steps \
                and (time_border is None or time.time() - global_start < time_border):
            print(f"-- Generation {generation}, population={len(pop)}, best time={best_fitness} --")
            if best_fitness == prev_best_fitness:
                plateau_steps += 1
            else:
                plateau_steps = 0
            prev_best_fitness = best_fitness

            # select individuals of next generation
            offspring = toolbox.select(pop, int(math.sqrt(len(pop))))
            # clone selected individuals
            # offspring = [toolbox.clone(ind) for ind in offspring]

            # operations for ORDER
            # crossover
            # take 2 individuals as input 1 modified individuals
            # take after 1: (1,3,5) and (2,4,6) and get pairs 1,2; 3,4; 5,6

            cur_generation = []

            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                if rand.random() < cxpb:
                    ind1, ind2 = toolbox.mate(child1[0], child2[0])
                    # add to population
                    cur_generation.append(wrap(ind1))
                    cur_generation.append(wrap(ind2))

            # mutation
            # take 1 individuals as input and return 1 individuals as output
            for mutant in offspring:
                if rand.random() < mutpb:
                    ind_order = toolbox.mutate(mutant[0][0])
                    ind = copy_chromosome(mutant[0])
                    ind = (ind_order[0], ind[1], ind[2])
                    # add to population
                    cur_generation.append(wrap(ind))

            # operations for RESOURCES
            # mutation
            # select types for mutation
            # numbers of changing types
            number_of_type_for_changing = rand.randint(1, len(worker_name2index) - 1)
            # workers type for changing(+1 means contractor 'resource')
            workers = rand.sample(range(len(worker_name2index) + 1), number_of_type_for_changing)

            # resources mutation
            for worker in workers:
                low = resources_border[0, worker] if worker != len(worker_name2index) else 0
                up = resources_border[1, worker] if worker != len(worker_name2index) else 0
                for mutant in offspring:
                    if rand.random() < mutpb_res:
                        ind = toolbox.mutate_resources(mutant[0], low=low, up=up, type_of_worker=worker)
                        # add to population
                        cur_generation.append(wrap(ind))

            # resource borders mutation
            for worker in workers:
                if worker == len(worker_name2index):
                    continue
                for mutant in offspring:
                    if rand.random() < mutpb_res:
                        ind = toolbox.mutate_resource_borders(mutant[0],
                                                              contractors_capacity=contractors_capacity,
                                                              resources_min_border=resources_min_border,
                                                              type_of_worker=worker)
                        # add to population
                        cur_generation.append(wrap(ind))

            # for the crossover, we use those types that did not participate
            # in the mutation(+1 means contractor 'resource')
            # workers_for_mate = list(set(list(range(len(worker_name2index) + 1))) - set(workers))
            # crossover
            # take 2 individuals as input 1 modified individuals

            workers = rand.sample(range(len(worker_name2index) + 1), number_of_type_for_changing)

            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                for ind_worker in workers:
                    # mate resources
                    if rand.random() < cxpb_res:
                        ind1, ind2 = toolbox.mate_resources(child1[0], child2[0], ind_worker)
                        # add to population
                        cur_generation.append(wrap(ind1))
                        cur_generation.append(wrap(ind2))

                    # mate resource borders
                    if rand.random() < cxpb_res:
                        if ind_worker == len(worker_name2index):
                            continue
                        ind1, ind2 = toolbox.mate_resource_borders(child1[0], child2[0], ind_worker)

                        # add to population
                        cur_generation.append(wrap(ind1))
                        cur_generation.append(wrap(ind2))

            evaluation_start = time.time()

            # Gather all the fitness in one list and print the stats
            invalid_ind = [ind for ind in cur_generation if toolbox.validate(ind[0])]
            # for each individual - evaluation
            # print(pool.map(lambda x: x + 2, range(10)))

            invalid_fit = fitness_f.evaluate([ind[0] for ind in invalid_ind])
            for fit, ind in zip(invalid_fit, invalid_ind):
                ind.fitness.values = [fit]
            evaluation_time += time.time() - evaluation_start

            # add mutant part of generation to offspring
            offspring.extend(invalid_ind)
            cur_generation.clear()

            if show_fitness_graph:
                _ftn = [f for f in fitness if not math.isinf(f)]
                if len(_ftn) > 0:
                    fitness_history.append(sum(_ftn) / len(_ftn))

            # pop_size = len(pop)
            # pop = [ind for ind in pop if valid(ind)]
            # print(f'----| Filtered out {pop_size - len(pop)} invalid individuals')

            # renewing population
            pop[:] = offspring
            hof.update(pop)

            best_fitness = hof[0].fitness.values[0]

            # best = hof[0]
            # fits = [ind.fitness.values[0] for ind in pop]
            # evaluation = chromosome_evaluation(best, index2node, resources_border, work_id2index, worker_name2index,
            #                                   parent2inseparable_son, agents)
            # print("fits: ", fits)
            # print(evaluation)
            generation += 1

        native.close()

        chromosome = hof[0][0]

        # assert that we have valid chromosome
        assert hof[0].fitness.values[0] != Time.inf()

        print(f'Final time: {hof[0].fitness.values[0]}')
        print(f'Generations processing took {(time.time() - start) * 1000} ms')
        print(f'Evaluation time: {evaluation_time * 1000}')
    else:
        native_start = time.time()
        chromosome = native.run_genetic(list([ind[0] for ind in pop]),
                                        mutate_order, mutate_order, mutate_resources, mutate_resources,
                                        mutate_resources, mutate_resources, selection_size)
        print(f'Native evaluated in {(time.time() - native_start) * 1000} ms')

    scheduled_works, schedule_start_time, timeline, order_nodes = toolbox.chromosome_to_schedule(chromosome,
                                                                                                 landscape=landscape,
                                                                                                 timeline=timeline)

    if show_fitness_graph:
        sns.lineplot(
            data=DataFrame.from_records([(generation * 4, v) for generation, v in enumerate(fitness_history)],
                                        columns=["Поколение", "Функция качества"]),
            x="Поколение",
            y="Функция качества",
            palette='r')
        plt.show()

    return {node.id: work for node, work in scheduled_works.items()}, schedule_start_time, timeline, order_nodes


def compare_individuals(first: tuple[ChromosomeType], second: tuple[ChromosomeType]):
    return (first[0][0] == second[0][0]).all() and (first[0][1] == second[0][1]).all()


def wrap(chromosome: ChromosomeType) -> Individual:
    """
    Created an individual from chromosome.
    """

    def ind_getter():
        return chromosome

    ind = initRepeat(Individual, ind_getter, n=1)
    ind.fitness.invalid_steps = 0
    return ind



from sampo.scheduler.generic import GenericScheduler


class DelegatingScheduler(GenericScheduler):
    """
    The layer between Generic Scheduler and end Scheduler.
    It's needed to change parametrization functions received from end Scheduler in a simple way.
    """

    def __init__(self, delegate: GenericScheduler):
        super().__init__(delegate.scheduler_type,
                         delegate.resource_optimizer,
                         delegate._timeline_type,
                         self.delegate_prioritization(delegate.prioritization),
                         self.delegate_optimize_resources(delegate.optimize_resources),
                         delegate.work_estimator)

    # noinspection PyMethodMayBeStatic
    def delegate_prioritization(self, prioritization):
        return prioritization

    # noinspection PyMethodMayBeStatic
    def delegate_optimize_resources(self, optimize_resources):
        return optimize_resources



class SchedulingPipelineError(Exception):
    """
    Raised when scheduling pipeline error occurred.

    This is a kind of 'IllegalStateException', e.g. raising this
    indicates that the corresponding scheduling pipeline come to incorrect internal state.
    """
    def __init__(self, message: str):
        super().__init__(message)



from abc import ABC, abstractmethod
from copy import deepcopy

from sampo.schemas.interval import IntervalGaussian
from sampo.schemas.resources import Resource, Material
from sampo.schemas.time import Time


class ResourceSupply(Resource, ABC):
    def __init__(self, id: str, name: str, count: int):
        super(ResourceSupply, self).__init__(id, name, count)

    @abstractmethod
    def get_available_resources(self) -> list[tuple[int, str]]:
        ...


class ResourceHolder(ResourceSupply):
    def __init__(self, id: str, name: str, productivity: IntervalGaussian, materials: list[Material]):
        super(ResourceHolder, self).__init__(id, name, int(productivity.mean))
        self._productivity = productivity
        self._materials = materials

    @property
    def productivity(self):
        return self._productivity

    def copy(self):
        return ResourceHolder(self.id, self.name, self.productivity, deepcopy(self._materials))

    def get_available_resources(self) -> list[tuple[int, str]]:
        return [(mat.count, mat.name) for mat in self._materials]


class Road(ResourceSupply):
    def __init__(self, id: str, name: str, throughput: IntervalGaussian):
        super(Road, self).__init__(id, name, int(throughput.mean))
        self._throughput = throughput

    @property
    def throughput(self):
        return self._throughput

    def copy(self):
        return Road(self.id, self.name, self.throughput)

    def get_available_resources(self) -> list[tuple[int, str]]:
        return []


class LandscapeConfiguration:
    def __init__(self, roads: list[Road] = [], holders: list[ResourceHolder] = []):
        self._roads = roads
        self._holders = holders

    def get_all_resources(self) -> list[ResourceSupply]:
        return self._roads + self._holders


class MaterialDelivery:
    def __init__(self, work_id: str):
        self.id = work_id
        self.delivery = {}

    def add_delivery(self, name: str, time: Time, count: int):
        material_delivery = self.delivery.get(name, None)
        if material_delivery is None:
            material_delivery = []
            self.delivery[name] = material_delivery
        material_delivery.append((time, count))



import os
from ctypes import *

from deap.base import Toolbox

from sampo.schemas.time import Time

native = True
try:
    from native import decodeEvaluationInfo
    from native import evaluate as evaluator
    from native import freeEvaluationInfo
    from native import runGenetic
except ImportError:
    print("Can't find native module; switching to default")
    decodeEvaluationInfo = lambda *args: args
    freeEvaluationInfo = lambda *args: args
    runGenetic = lambda *args: args
    evaluator = None
    native = False


from sampo.scheduler.genetic.converter import ChromosomeType
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.collections_util import reverse_dictionary


class NativeWrapper:
    def __init__(self,
                 toolbox: Toolbox,
                 wg: WorkGraph,
                 contractors: list[Contractor],
                 worker_name2index: dict[str, int],
                 worker_pool_indices: dict[int, dict[int, Worker]],
                 parents: dict[int, list[int]],
                 time_estimator: WorkTimeEstimator):
        self.native = native
        if not native:
            def fit(chromosome: ChromosomeType) -> int:
                if toolbox.validate(chromosome):
                    sworks = toolbox.chromosome_to_schedule(chromosome)[0]
                    return max([swork.finish_time for swork in sworks.values()]).value
                else:
                    return Time.inf()
            self.evaluator = lambda _, chromosomes: [fit(chromosome) for chromosome in chromosomes]
            self._cache = None
            return

        # the outer numeration. Begins with inseparable heads, continuous with tails.
        numeration: dict[int, GraphNode] = {i: node for i, node in
                                            enumerate(filter(lambda node: not node.is_inseparable_son(), wg.nodes))}
        heads_count = len(numeration)
        for i, node in enumerate([node for node in wg.nodes if node.is_inseparable_son()]):
            numeration[heads_count + i] = node
        rev_numeration = reverse_dictionary(numeration)

        # TODO remove assignment unuseful info to self

        self.numeration = numeration
        # for each vertex index store list of parents' indices
        self.parents = [[rev_numeration[p] for p in numeration[index].parents] for index in range(wg.vertex_count)]
        head_parents = [parents[i] for i in range(len(parents))]
        # for each vertex index store list of whole it's inseparable chain indices
        self.inseparables = [[rev_numeration[p] for p in numeration[index].get_inseparable_chain_with_self()]
                             for index in range(wg.vertex_count)]
        # contractors' workers matrix. If contractor can't supply given type of worker, 0 should be passed
        self.workers = [[0 for _ in range(len(worker_name2index))] for _ in contractors]
        for i, contractor in enumerate(contractors):
            for worker in contractor.workers.values():
                self.workers[i][worker_name2index[worker.name]] = worker.count

        min_req = [[] for _ in range(len(numeration))]  # np.zeros((len(numeration), len(worker_pool_indices)))
        max_req = [[] for _ in range(len(numeration))]  # np.zeros((len(numeration), len(worker_pool_indices)))
        for work_index, node in numeration.items():
            cur_min_req = [0 for _ in worker_name2index]
            cur_max_req = [0 for _ in worker_name2index]
            for req in node.work_unit.worker_reqs:
                worker_index = worker_name2index[req.kind]
                cur_min_req[worker_index] = req.min_count
                cur_max_req[worker_index] = req.max_count
            min_req[work_index] = cur_min_req
            max_req[work_index] = cur_max_req

        volume = [node.work_unit.volume for node in numeration.values()]

        id2work = [numeration[i].work_unit.id for i in range(len(numeration))]
        id2worker_name = reverse_dictionary(worker_name2index)
        id2res = [id2worker_name[i] for i in range(len(id2worker_name))]

        self.totalWorksCount = wg.vertex_count
        self.time_estimator = time_estimator
        self.worker_pool_indices = worker_pool_indices
        self._current_chromosomes = None

        self.evaluator = evaluator

        # os.add_dll_directory('C:\\Users\\Quarter\\PycharmProjects\\sampo\\tests')
        # os.add_dll_directory('C:\\Users\\Quarter\\PycharmProjects\\sampo')
        # os.chdir('C:\\Users\\Quarter\\PycharmProjects\\sampo\\tests')

        import ctypes.util

        name = ctypes.util.find_library('C:\\Users\\Quarter\\PycharmProjects\\sampo\\tests\\native.dll')
        lib = ctypes.WinDLL(name)
        # lib = CDLL(r'C:\Users\Quarter\PycharmProjects\sampo\tests\native.dll')

        # preparing C++ cache
        self._cache = decodeEvaluationInfo(self, self.parents, head_parents, self.inseparables, self.workers,
                                           self.totalWorksCount, False, True, volume, min_req, max_req, id2work, id2res)

    def calculate_working_time(self, chromosome_ind: int, team_target: int, work: int) -> int:
        team = self._current_chromosomes[chromosome_ind][1][team_target]
        workers = [self.worker_pool_indices[worker_index][team[len(self.workers[0])]]
                   .copy().with_count(team[worker_index])
                   for worker_index in range(len(self.workers[0]))
                   if team[worker_index] > 0]
        return self.time_estimator.estimate_time(self.numeration[work].work_unit, workers).value

    def evaluate(self, chromosomes: list[ChromosomeType]):
        self._current_chromosomes = chromosomes
        return self.evaluator(self._cache, chromosomes)

    def run_genetic(self, chromosomes: list[ChromosomeType],
                    mutate_order, mate_order, mutate_resources, mate_resources,
                    mutate_contractors, mate_contractors, selection_size):
        self._current_chromosomes = chromosomes
        return runGenetic(self._cache, chromosomes, mutate_order, mutate_resources, mutate_contractors,
                          mate_order, mate_resources, mate_contractors, selection_size)

    def close(self):
        freeEvaluationInfo(self._cache)



class NoSufficientContractorError(Exception):
    """
    Raise when contractor error occurred.

    It indicates when the contractors have not sufficient resources to perform schedule.
    """
    def __init__(self, message):
        super().__init__(message)


class NotEnoughMaterialsInDepots(Exception):

    def __init__(self, message):
        super().__init__(message)


class NoAvailableResources(Exception):

    def __init__(self, message):
        super().__init__(message)



from collections import deque
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property, cache
from typing import Union, Optional

import numpy as np
from scipy.sparse import dok_matrix

from sampo.schemas.serializable import JSONSerializable, T, JS
from sampo.schemas.works import WorkUnit


class EdgeType(Enum):
    """
    Class to define a certain type of edge in graph
    """
    InseparableFinishStart = 'IFS'
    LagFinishStart = 'FFS'
    StartStart = 'SS'
    FinishFinish = 'FF'
    FinishStart = 'FS'
    StartFinish = 'SF'

    @staticmethod
    def is_dependency(edge) -> bool:
        if edge == '-1':  # ... no comments
            return True
        if isinstance(edge, EdgeType):
            edge = edge.value
        return edge in ('FS', 'IFS', 'FFS')


@dataclass
class GraphEdge:
    """
    The edge of graph with start and finish vertexes
    """
    start: 'GraphNode'
    finish: 'GraphNode'
    lag: float | None = 0
    type: EdgeType | None = None


class GraphNode(JSONSerializable['GraphNode']):
    """
    Class to describe Node in graph
    """
    _work_unit: WorkUnit
    _parent_edges: list[GraphEdge]
    _children_edges: list[GraphEdge]

    def __init__(self, work_unit: WorkUnit,
                 parent_works: Union[list['GraphNode'], list[tuple['GraphNode', float, EdgeType]]]):
        self._work_unit = work_unit
        self._parent_edges = []
        self.add_parents(parent_works)
        self._children_edges = []

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return self.id

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        return self._work_unit._serialize(), \
            [(e.start.id, e.lag, e.type.value) for e in self._parent_edges]

    def __setstate__(self, state):
        # custom method to avoid calling __hash__() on GraphNode objects
        s_work_unit, s_parent_edges = state
        self.__init__(WorkUnit._deserialize(s_work_unit),
                      s_parent_edges)
        # self._work_unit = representation['work_unit']
        # self._parent_edges = [GraphEdge(*e) for e in representation['parent_edges']]

    def _serialize(self) -> T:
        return {
            'work_unit': self._work_unit._serialize(),
            'parent_edges': [(e.start.id, e.lag, e.type.value) for e in self._parent_edges],
            # 'child_edges': [(e.finish.work_unit.id, e.lag, e.type.value) for e in self._children_edges]
        }

    @classmethod
    def _deserialize(cls, representation: T) -> dict:
        representation['work_unit'] = WorkUnit._deserialize(representation['work_unit'])
        representation['parent_edges'] = [(e[0], e[1], EdgeType(e[2])) for e in representation['parent_edges']]
        # representation['child_edges'] = [(e[0], e[1], EdgeType(e[2])) for e in representation['child_edges']]
        return representation

    def update_work_unit(self, work_unit: WorkUnit) -> None:
        self._work_unit = work_unit

    def add_parents(self, parent_works: list['GraphNode'] or list[tuple['GraphNode', float, EdgeType]]) -> None:
        """
        Two-sided linking of successors and predecessors

        :param parent_works: list of parent works
        """
        edges: list[GraphEdge] = []
        if len(parent_works) > 0 and isinstance(parent_works[0], GraphNode):
            edges = [GraphEdge(p, self, -1, EdgeType.FinishStart) for p in parent_works]
        elif len(parent_works) > 0 and isinstance(parent_works[0], tuple):
            edges = [GraphEdge(p, self, lag, edge_type) for p, lag, edge_type in parent_works]

        for i, parent in enumerate(parent_works):
            parent: GraphNode = parent[0] if isinstance(parent, tuple) else parent
            parent._add_child_edge(edges[i])
            parent.invalidate_children_cache()
        self._parent_edges += edges
        self.invalidate_parents_cache()

    def invalidate_parents_cache(self):
        self.__dict__.pop('parents', None)
        self.__dict__.pop('parents_set', None)
        self.__dict__.pop('inseparable_parent', None)
        self.__dict__.pop('inseparable_son', None)
        self.__dict__.pop('get_inseparable_chain', None)

    def invalidate_children_cache(self):
        self.__dict__.pop('children', None)
        self.__dict__.pop('children_set', None)
        self.__dict__.pop('inseparable_parent', None)
        self.__dict__.pop('inseparable_son', None)
        self.__dict__.pop('get_inseparable_chain', None)

    def is_inseparable_parent(self) -> bool:
        return self.inseparable_son is not None

    def is_inseparable_son(self) -> bool:
        return self.inseparable_parent is not None

    def traverse_children(self, topologically: bool = False):
        """
        DFS from current vertex to down
        :param topologically: is DFS need to go in topologically way
        :return:
        """
        visited_vertexes = set()
        vertexes_to_visit = deque([self])
        while len(vertexes_to_visit) > 0:
            v = vertexes_to_visit.popleft()
            if topologically and any(p.start not in visited_vertexes for p in v._parent_edges):
                vertexes_to_visit.append(v)
                continue
            if v not in visited_vertexes:
                visited_vertexes.add(v)
                vertexes_to_visit.extend([p.finish for p in v._children_edges])
                yield v

    @cached_property
    # @property
    def inseparable_son(self) -> Optional['GraphNode']:
        """
        Return inseparable son (amount of inseparable sons at most 1)
        :return: inseparable son
        """
        inseparable_children = list([x.finish for x in self._children_edges
                                     if x.type == EdgeType.InseparableFinishStart])
        return inseparable_children[0] if inseparable_children else None

    @cached_property
    # @property
    def inseparable_parent(self) -> Optional['GraphNode']:
        """
        Return predecessor of current vertex in inseparable chain
        :return: inseparable parent
        """
        inseparable_parents = list([x.start for x in self._parent_edges if x.type == EdgeType.InseparableFinishStart])
        return inseparable_parents[0] if inseparable_parents else None

    @cached_property
    # @property
    def parents(self) -> list['GraphNode']:
        """
        Return list of predecessors of current vertex
        :return: list of parents
        """
        return list([edge.start for edge in self.edges_to if EdgeType.is_dependency(edge.type)])

    @cached_property
    # @property
    def parents_set(self) -> set['GraphNode']:
        """
        Return unique predecessors of current vertex
        :return: set of parents
        """
        return set(self.parents)

    @cached_property
    # @property
    def children(self) -> list['GraphNode']:
        """
        Return list of successors of current vertex
        :return: list of children
        """
        return list([edge.finish for edge in self.edges_from if EdgeType.is_dependency(edge.type)])

    @cached_property
    # @property
    def children_set(self) -> set['GraphNode']:
        """
        Return unique successors of current vertex
        :return: set of children
        """
        return set(self.children)

    @cached_property
    def neighbors(self):
        """
        Get all edges that have types SS with current vertex
        :return: list of neighbours
        """
        return list([edge.start for edge in self._parent_edges if edge.type == EdgeType.StartStart])

    @property
    def edges_to(self) -> list[GraphEdge]:
        return self._parent_edges

    @property
    def edges_from(self) -> list[GraphEdge]:
        """
        Return all successors of current vertex
        :return: list of successors
        """
        return self._children_edges

    @property
    def work_unit(self) -> WorkUnit:
        return self._work_unit

    @property
    def id(self) -> str:
        return self.work_unit.id

    @cache
    def get_inseparable_chain(self) -> Optional[list['GraphNode']]:
        """
        Gets an ordered list of whole chain of nodes, connected with edges of type INSEPARABLE_FINISH_START =
        'INSEPARABLE',
        IF self NODE IS THE START NODE OF SUCH CHAIN. Otherwise, None.

        :return: list of GraphNode or None
        """
        return [self] + self._get_inseparable_children() \
            if bool(self.inseparable_son) and not bool(self.inseparable_parent) \
            else None

    def get_inseparable_chain_with_self(self) -> list['GraphNode']:
        """
        Gets an ordered list of whole chain of nodes, connected with edges of type INSEPARABLE_FINISH_START =
        'INSEPARABLE'.

        :return: list of `inseparable chain` with starting node
        """
        return self.get_inseparable_chain() if self.get_inseparable_chain() else [self]

    def _get_inseparable_children(self) -> list['GraphNode']:
        """
        Recursively gets a child, connected with INSEPARABLE_FINISH_START edge, its inseparable child, etc.
        As any node may contain an inseparable connection with only one of its children, there is no need to choose.
        If no children are connected inseparably, returns None.

        :return: list[GraphNode]. Empty, if there is no inseparable children
        """
        inseparable_child = self.inseparable_son
        return [inseparable_child] + inseparable_child._get_inseparable_children() \
            if inseparable_child \
            else []

    def _add_child_edge(self, child: GraphEdge):
        """
        Append new edge with child

        :param child:
        :return: current graph node
        """
        self._children_edges.append(child)


GraphNodeDict = dict[str, GraphNode]


# TODO Make property for list of GraphEdges??
@dataclass
class WorkGraph(JSONSerializable['WorkGraph']):
    """
    Class to describe graph of works
    """
    # service vertexes
    start: GraphNode
    finish: GraphNode

    # list of works (i.e. GraphNode)
    nodes: list[GraphNode] = field(init=False)
    adj_matrix: dok_matrix = field(init=False)
    dict_nodes: GraphNodeDict = field(init=False)
    vertex_count: int = field(init=False)


    def __post_init__(self) -> None:
        self.reinit()

    def reinit(self):
        ordered_nodes, adj_matrix, dict_nodes = self._to_adj_matrix()
        # To avoid field set of frozen instance errors
        object.__setattr__(self, 'nodes', ordered_nodes)
        object.__setattr__(self, 'adj_matrix', adj_matrix)
        object.__setattr__(self, 'dict_nodes', dict_nodes)
        object.__setattr__(self, 'vertex_count', len(ordered_nodes))

    def __hash__(self):
        return hash(self.start) + 17 * hash(self.finish)

    def __getitem__(self, item: str) -> GraphNode:
        return self.dict_nodes[item]

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        representation = self._serialize()
        representation['start'] = self.start.id
        representation['finish'] = self.start.id
        return representation

    def __setstate__(self, state):
        # custom method to avoid calling __hash__() on GraphNode objects
        deserialized = self._deserialize(state)
        object.__setattr__(self, 'start', deserialized.start)
        object.__setattr__(self, 'finish', deserialized.finish)
        self.__post_init__()

    def _serialize(self) -> T:
        return {
            'nodes': [graph_node._serialize() for graph_node in self.nodes]
        }

    @classmethod
    def _deserialize(cls, representation: T) -> JS:
        serialized_nodes = [GraphNode._deserialize(node) for node in representation['nodes']]
        assert not serialized_nodes[0]['parent_edges']
        start_id, finish_id = (serialized_nodes[i]['work_unit'].id for i in (0, -1))

        nodes_dict = dict()
        for node_info in serialized_nodes:
            wu, parent_info = (node_info[member] for member in ('work_unit', 'parent_edges'))
            graph_node = GraphNode(wu, [(nodes_dict[p_id], p_lag, p_type) for p_id, p_lag, p_type in parent_info])
            nodes_dict[wu.id] = graph_node

        return WorkGraph(nodes_dict[start_id], nodes_dict[finish_id])

    # TODO: Check that adj matrix is really need
    def _to_adj_matrix(self) -> tuple[list[GraphNode], dok_matrix, dict[str, GraphNode]]:
        """
        Build adjacency matrix from current graph
        """
        ordered_nodes: list[GraphNode] = list(self.start.traverse_children(topologically=True))
        node2ind: dict[GraphNode, int] = {
            v: i for i, v in enumerate(ordered_nodes)
        }
        id2node: dict[str, GraphNode] = {node.id: node for node in node2ind.keys()}
        adj_mx = dok_matrix((len(node2ind), len(node2ind)), dtype=np.short)
        weight = 0
        for v, i in node2ind.items():
            for child in v.children:
                c_i = node2ind[child]
                weight = max((w_req.volume for w_req in v.work_unit.worker_reqs), default=0.000001)
                adj_mx[i, c_i] = weight

        return ordered_nodes, adj_mx, id2node



from sampo.schemas.graph import WorkGraph
from sampo.structurator.graph_insertion import prepare_work_graph_copy


def work_graph_ids_simplification(wg: WorkGraph, id_offset: int = 0) -> WorkGraph:
    """
    Creates a new WorkGraph with simplified numeric ids (numeric ids are converted to a string)
    :param wg: original WorkGraph
    :param id_offset: start for numbering new ids
    :return: new WorkGraph with numeric ids
    """
    nodes, old_to_new_ids = prepare_work_graph_copy(wg, [], use_ids_simplification=True, id_offset=id_offset)
    start = nodes[old_to_new_ids[wg.start.id]]
    finish = nodes[old_to_new_ids[wg.finish.id]]
    return WorkGraph(start, finish)



import time
from typing import Union, Callable

from sampo.scheduler.base import SchedulerType, Scheduler
from sampo.scheduler.genetic.base import GeneticScheduler
from sampo.scheduler.heft.base import HEFTBetweenScheduler
from sampo.scheduler.heft.base import HEFTScheduler
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.topological.base import TopologicalScheduler
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.schedule import Schedule
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.resource_cost import schedule_cost


def get_scheduler_ctor(scheduling_algorithm_type: SchedulerType) \
        -> Callable[[SchedulerType, ResourceOptimizer, WorkTimeEstimator], Scheduler]:
    if scheduling_algorithm_type is SchedulerType.HEFTAddBetween:
        return HEFTBetweenScheduler
    if scheduling_algorithm_type is SchedulerType.HEFTAddEnd:
        return HEFTScheduler
    if scheduling_algorithm_type is SchedulerType.Topological:
        return TopologicalScheduler
    return GeneticScheduler


def generate_schedule(scheduling_algorithm_type: SchedulerType,
                      work_time_estimator: WorkTimeEstimator,
                      work_graph: WorkGraph,
                      contractors: Union[Contractor, list[Contractor]],
                      validate_schedule: bool,
                      landscape: LandscapeConfiguration = LandscapeConfiguration()) -> Schedule:
    scheduler = get_scheduler_ctor(scheduling_algorithm_type)(work_estimator=work_time_estimator)
    start_time = time.time()
    if isinstance(scheduler, GeneticScheduler):
        scheduler.set_use_multiprocessing(n_cpu=4)

    schedule = scheduler.schedule(work_graph,
                                  [contractors] if isinstance(contractors, Contractor) else contractors,
                                  validate=validate_schedule, landscape=landscape)

    print(f'Time: {(time.time() - start_time) * 1000} ms')
    print(f'Cost: {schedule_cost(schedule)}')
    return schedule



from sampo.schemas.contractor import Contractor
from sampo.schemas.schedule import Schedule


def get_total_resource_usage(schedule: Schedule, contractors: list[Contractor]) -> dict[str, list[int]]:
    resources = list(set(res.name for contractor in contractors for res in contractor.workers.values()))
    usage = {res: [0 for _ in range(schedule.execution_time.value + 1)] for res in resources}

    for work in schedule.works:
        for worker in work.workers:
            for day in range(work.start_time.value, work.finish_time.value + 1):
                usage[worker.name][day] += worker.count
    return usage


def get_peak_resource_usage(schedule: Schedule, contractors: list[Contractor]) -> dict[str, int]:
    return {res: max(res_usage) for res, res_usage in get_total_resource_usage(schedule, contractors).items()}


def get_absolute_peak_resource_usage(schedule: Schedule, contractors: list[Contractor]) -> int:
    return sum(get_peak_resource_usage(schedule, contractors).values())



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional
from uuid import uuid4

from sampo.schemas.resources import Material
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time

# Used for max_count in the demand, if it is not specified during initialization WorkerReq
DEFAULT_MAX_COUNT = 100


class BaseReq(AutoJSONSerializable['BaseReq'], ABC):
    """
    A class summarizing any requirements for the work to be performed related to renewable and non-renewable
    resources, infrastructure requirements, etc.
    """
    @property
    @abstractmethod
    def name(self) -> str:
        """
        Returns the name of the claim if it exists, e.g. 'dig work claim'.

        :return name: the name of req
        """
        ...


@dataclass(frozen=True)
class WorkerReq(BaseReq):
    """
    Requirements related to renewable human resources

    :param kind: type of resource/profession
    :param volume: volume of work in time units
    :param min_count: minimum number of employees needed to perform the work
    :param max_count: maximum allowable number of employees performing work
    :param name: the name of this requirement
    """
    kind: str
    volume: Time
    min_count: Optional[int] = 1
    max_count: Optional[int] = DEFAULT_MAX_COUNT
    name: Optional[str] = ''

    def scale_all(self, scalar: float, new_name: Optional[str] = '') -> 'WorkerReq':
        """
        The function scales the requirement to the size of the work including the total
        volume and the maximum number of personnel involved.

        :param scalar: scalar for multiplication
        :param new_name: name for new req
        :return new_req: new object with new volume of the work and extended max_count_commands
        """
        max_count = max(round(self.max_count * scalar), self.min_count)
        new_req = WorkerReq(self.kind, self.volume * scalar, self.min_count, max_count, new_name or self.name)
        return new_req

    def scale_volume(self, scalar: float, new_name: Optional[str] = None) -> 'WorkerReq':
        """
        The function scales only volume of the work for the requirement.

        :param scalar: scalar for multiplication
        :param new_name: name for new req
        :return new_req: new object with new volume of the work.
        """
        new_req = WorkerReq(self.kind, self.volume * scalar, self.min_count, self.max_count, new_name or self.name)
        return new_req


@dataclass(frozen=True)
class EquipmentReq(BaseReq):
    """
    Requirements for renewable non-human resources: equipment, trucks, machines, etc

    :param kind: type of resource/profession
    :param name: the name of this requirement
    """
    kind: str
    name: Optional[str] = None


@dataclass(frozen=True)
class MaterialReq(BaseReq):
    """
    Requirements for non-renewable materials: consumables, spare parts, construction materials

    :param kind: type of resource/profession
    :param name: the name of this requirement
    """
    kind: str
    count: int
    # need_start: int TODO Implement handling this
    name: Optional[str] = None

    def material(self):
        return Material(str(uuid4()), self.kind, self.count)


@dataclass(frozen=True)
class ConstructionObjectReq(BaseReq):
    """
    Requirements for infrastructure and the construction of other facilities: electricity, pipelines, roads, etc

    :param kind: type of resource/profession
    :param name: the name of this requirement
    """
    kind: str
    name: Optional[str] = None
